/**
  Pulse-Width-Modulation (PWM) driver for PIC16F chips.
 
  Orion Sky Lawlor, olawlor@acm.org, 2007-05-21 (Public Domain)
*/
#include "pickit.h"
#include "serial.h"


#ifndef CHIP_B /* if we're not B */
#  define CHIP_A 1 /* then we're A */
#endif


__IDLOC(1);
__CONFIG(INTIO & WDTDIS & MCLRDIS & BORDIS & UNPROTECT & PWRTEN);

/** Received a bad serial data byte (user routine) */
void SerialRecvBad(void) {
	SerialRecvData='E';
}

/* Busy-wait for this fraction of a millisecond.
  FIXME: using timer here would use less power and generate less heat
*/
void busywait_subms(byte fract) {
	// Wait for one millisecond (depends on compiler, optimizer)
	byte c;
	for (c=fract;c!=0;c--) { NOP(); }
}

/* Run this GPIO line (expressed as a bit mask)
   at this fractional power.  0 sends a 1ms pulse,
   255 sends a 2ms pulse.  This is the standard servo
   Pulse-Width-Modulation protocol.
*/
void fire_pwm(byte fract, byte pin) {
	GPIO |= pin; /* turn pin on */
	busywait_subms(250); /* wait 1ms always */
	busywait_subms(fract); /* wait variable delay */
	GPIO &= ~pin; /* turn pin off again */
}
/* like fire_pwm, but uses a pin on PORTC for 14-pin controllers. */
void fire_pwmc(byte fract, byte old, byte pin) {
	PORTC = old | pin; /* turn pin on */
	busywait_subms(250); /* wait 1ms always */
	busywait_subms(fract); /* wait variable delay */
	PORTC = old & ~pin; /* turn pin off again */
}

void main(void) {
	load_osccal();
	TRISIO = 1<<SerialInputBit; /* serial input pin, all else output */
#if CHIP_B
	TRISIO |= 1<<SerialOutputBit; /* don't mess with A's outputs */
#endif
	TRISC = 0; /* all six port-c pins are outputs */
	
	CMCON = 0x07; /* comparator off */
	WPU=0; /* disable weak pull-up on all pins*/
        ANSEL = 0x00; /* disable A/D */
        ADCON0 = 0x00;
	
        OPTION = 0x80; /* disable weak pull-up globally */
        TMR0 = 0;
	
	/* Turn on interrupts on GPIO0 */
	IOCB=0x01; /* interrupt on pin 0 change */
	GPIO = 0x0; /* clear output pins */
	GPIF=0; /* clear interrupt flag */
	GPIE=1; /* enable GPIO interrupts */
	
	GIE=0; /* turn off interrupts */
	busywait(250); /* wait for lines to settle before sending */
	
	/* Main microcontroller control loop: receive command, execute it. */
	while (1) {
		byte spikes,
			f0,f1,f2,f3,f4,f5,
			b0,b1,b2,b3,b4,b5,
			exec,
			rpt=0;
#if CHIP_A 
		// Send "we're A-OK, ready to receive 12 PWM inputs"
		SerialSendByte(0xA2);
#endif
		// Synchronize and receive the start command
		do {
			SerialRecvByte(); exec=SerialRecvData;
		} while (exec!=0xEC);
		
		// Receive the spike values
		SerialRecvByte(); spikes=SerialRecvData;
		
		// Receive the 12 PWM values
		SerialRecvByte(); f0=SerialRecvData;
		SerialRecvByte(); f1=SerialRecvData;
		SerialRecvByte(); f2=SerialRecvData;
		SerialRecvByte(); f3=SerialRecvData;
		SerialRecvByte(); f4=SerialRecvData;
		SerialRecvByte(); f5=SerialRecvData;
		
		SerialRecvByte(); b0=SerialRecvData;
		SerialRecvByte(); b1=SerialRecvData;
		SerialRecvByte(); b2=SerialRecvData;
		SerialRecvByte(); b3=SerialRecvData;
		SerialRecvByte(); b4=SerialRecvData;
		SerialRecvByte(); b5=SerialRecvData;
		
		// Transmit those values (possibly for a while)
		// for (rpt=100;rpt!=0;rpt--) 
		{
			unsigned char portc_spikes;
#if CHIP_B /* B is firing its own PWMs.  
	FIXME: would arrays of values work here? */
			f0=b0; f1=b1; f2=b2; f3=b3; f4=b4; f5=b5;
			spikes=spikes>>4;
#endif
		/* set the spike values.  PWMs are all off at this point,
		   so just force the PWM pins to zeros. */
			PORTC = portc_spikes = (spikes&0x7)<<3;
			
		/* send PWM values (could trivially overlap these too...) */
			fire_pwm(f0,1<<0);
			fire_pwm(f1,1<<1);
			fire_pwm(f2,1<<2);
			fire_pwmc(f3,portc_spikes,1<<0); /* these PWMs are on port C */
			fire_pwmc(f4,portc_spikes,1<<1);
			fire_pwmc(f5,portc_spikes,1<<2);
		}
	}
}
